from pokeman.utils.custom_abc import ABCMeta, abstract_attribute

import logging

LOGGER = logging.getLogger(__name__)


class AbstractExchange(metaclass=ABCMeta):
    """
    Abstract base class for Exchange object.
    """

    @abstract_attribute
    def exchange_name(self):
        pass

    @abstract_attribute
    def exchange_type(self):
        pass

    @abstract_attribute
    def durable(self):
        pass

    @abstract_attribute
    def auto_delete(self):
        pass

    @abstract_attribute
    def specific_poker(self):
        pass


class AbstractRoutingKey(metaclass=ABCMeta):
    """
    Abstract base class for RoutingKey object.
    """

    @abstract_attribute
    def key(self):
        pass


# TODO: Add arguments: arguments, callback for declaring and binding purposes
class AbstractQueue(metaclass=ABCMeta):
    """
    Abstract base class for Queue object.
    """

    @abstract_attribute
    def queue_name(self):
        pass

    @abstract_attribute
    def exchange(self):
        pass

    @abstract_attribute
    def routing_key(self):
        pass

    @abstract_attribute
    def exclusive(self):
        pass

    @abstract_attribute
    def durable(self):
        pass

    @abstract_attribute
    def auto_delete(self):
        pass

    @abstract_attribute
    def specific_poker(self):
        pass


# TODO: Add arguments: internal, arguments, callback
# TODO: Build custom passive handler
class Exchange(AbstractExchange):
    """
    Exchange object.
    """
    _SENTINEL = Ellipsis

    def __init__(self,
                 exchange_name=_SENTINEL,
                 exchange_type=_SENTINEL,
                 durable=_SENTINEL,
                 auto_delete=_SENTINEL,
                 specific_poker=_SENTINEL):
        """
        This method initialized the exchange object.

        :param exchange_name: The provided exchange name.
        :type exchange_name: str

        :param exchange_type: The provided exchange type.
        :type exchange_type: str

        :param durable: The provided durable parameter.
        :type durable: bool

        :param auto_delete: The provided auto_delete parameter.
        :type auto_delete: bool

        :param specific_poker: The provided specific poker id to handle the object.
        :type specific_poker: pokeman.Pokeman
        """
        from pokeman.amqp_resources.heapq import ResourceHeapQ  # Avoid circular import
        self.exchange_name = self._exchange_name_setter(exchange_name=exchange_name)
        self.exchange_type = self._exchange_type_setter(exchange_type=exchange_type)
        self.durable = self._durable_setter(durable=durable)
        self.auto_delete = self._auto_delete_setter(auto_delete=auto_delete)
        self.specific_poker = self._specific_poker_setter(specific_poker=specific_poker)
        ResourceHeapQ.add_resource(resource=self, specific_poker=self.specific_poker)

    def _exchange_name_setter(self, exchange_name):
        if exchange_name is self._SENTINEL:
            return 'pokeman_exchange'
        else:
            return exchange_name

    def _exchange_type_setter(self, exchange_type):
        if exchange_type is self._SENTINEL:
            return 'direct'
        else:
            return exchange_type

    def _durable_setter(self, durable):
        if durable is self._SENTINEL:
            return True
        else:
            return durable

    def _auto_delete_setter(self, auto_delete):
        if auto_delete is self._SENTINEL:
            return False
        else:
            return auto_delete

    def _specific_poker_setter(self, specific_poker):
        if specific_poker is self._SENTINEL:
            return None
        elif specific_poker is None:
            return None
        else:
            return specific_poker.POKER_ID


class RoutingKey(AbstractRoutingKey):
    """
    Routing key object.
    """
    _SENTINEL = Ellipsis

    def __init__(self, key=_SENTINEL):
        """
        This method initializes the routing key object.

        :param key: The provided routing key.
        :type key: str
        """
        self.key = self._routing_key_setter(key=key)

    def _routing_key_setter(self, key):
        if key is self._SENTINEL:
            return None
        else:
            return key


class Queue(AbstractQueue):
    """
    Queue object
    """
    _SENTINEL = Ellipsis

    def __init__(self,
                 queue_name=_SENTINEL,
                 exchange=_SENTINEL,
                 routing_key=_SENTINEL,
                 exclusive=_SENTINEL,
                 durable=_SENTINEL,
                 auto_delete=_SENTINEL,
                 specific_poker=_SENTINEL
                 ):
        """
        This method initializes the queue object.

        :param queue_name: The provided queue name.
        :type queue_name: str

        :param exchange: The provided exchange.
        :type exchange: Exchange

        :param routing_key: The provided routing key.
        :type routing_key: RoutingKey

        :param exclusive: The provided exclusive parameter.
        :type exclusive: bool

        :param auto_delete: The provided auto delete parameter.
        :type auto_delete: bool

        :param specific_poker: The provided specific poker id to handle the object.
        :type specific_poker: pokeman.Pokeman
        """
        from pokeman.amqp_resources.heapq import ResourceHeapQ
        self._exchange_converter = lambda: Exchange(
            specific_poker=self._specific_poker_setter(
                specific_poker=specific_poker,
                lambda_call=True
            )).exchange_name if exchange is self._SENTINEL else exchange.exchange_name
        self._routing_key_converter = lambda: RoutingKey().key if routing_key is self._SENTINEL else routing_key.key
        self.queue_name = self._queue_name_setter(queue_name=queue_name)
        self.exchange = self._exchange_setter(exchange=self._exchange_converter())
        self.routing_key = self._routing_key_setter(routing_key=self._routing_key_converter())
        self.exclusive = self._exclusive_setter(exclusive=exclusive)
        self.durable = self._durable_setter(durable=durable)
        self.auto_delete = self._auto_delete_setter(auto_delete=auto_delete)
        self.specific_poker = self._specific_poker_setter(specific_poker=specific_poker)
        ResourceHeapQ.add_resource(resource=self, specific_poker=self.specific_poker)

    def _queue_name_setter(self, queue_name):
        if queue_name is self._SENTINEL:
            return "pokeman_queue"
        else:
            return queue_name

    def _exchange_setter(self, exchange):
        return exchange

    def _routing_key_setter(self, routing_key):
        return routing_key

    def _exclusive_setter(self, exclusive):
        if exclusive is self._SENTINEL:
            return False
        else:
            return exclusive

    def _durable_setter(self, durable):
        if durable is self._SENTINEL:
            return True
        else:
            return durable

    def _auto_delete_setter(self, auto_delete):
        if auto_delete is self._SENTINEL:
            return False
        else:
            return auto_delete

    def _specific_poker_setter(self, specific_poker, lambda_call=False):
        if specific_poker is self._SENTINEL:
            return None
        elif lambda_call is False:
            return specific_poker.POKER_ID
        elif lambda_call is True and specific_poker == self.specific_poker:
            return None
        elif lambda_call is True and specific_poker != self.specific_poker:
            return specific_poker
        else:
            raise ValueError('Unresolvable specific poker parameter')



from enum import Enum


class TaskStatus(Enum):
    """
    The available task status variables.
    """

    PENDING = 'PENDING'
    SUCCESS = 'SUCCESS'
    ON_HOLD = 'ON_HOLD'
    FAILED = 'FAILED'



import unittest
from unittest import mock
import pika

from tests import TestAttributes as TA
from pokeman.composite import connection
from pokeman.composite.config import configs


class ConnectionTypetests(unittest.TestCase):
    def setUp(self):
        self.ConnectionType = connection.ConnectionType

    def test_base_name(self):
        self.assertEqual(self.ConnectionType.__name__, 'ConnectionType')

    def test_enums(self):
        self.assertEqual(self.ConnectionType.URL.name, 'URL')
        self.assertEqual(self.ConnectionType.HOST.name, 'HOST')


class AbstractConnectionTests(unittest.TestCase):
    def test_base_name(self):
        self.assertEqual(connection.AbstractConnection.__name__, 'AbstractConnection')

    @mock.patch.object(connection.AbstractConnection, '__abstractmethods__', set())
    def test_base_initialization(self):
        self.basic_config = configs.BasicConfig()
        self.abc_connection = connection.AbstractConnection(
            connstr='amqp://username:password@localhost:5672',
            method=connection.ConnectionType.URL,
            config=self.basic_config
        )
        self.assertEqual(self.abc_connection._connection_string, 'amqp://username:password@localhost:5672')
        self.assertEqual(self.abc_connection._method, connection.ConnectionType.URL)
        self.assertIsInstance(self.abc_connection._config, configs.BasicConfig)


class BasicConnectionTests(unittest.TestCase):
    class ConnectionMock:
        def __init__(self, status):
            self.status = status

        def close(self):
            if self.status == 'success':
                self.is_closed = True
            else:
                self.is_closed = False

    def setUp(self):
        self.basic_config = configs.BasicConfig()
        self.basic_connection = connection.BasicConnection(
            connstr='amqp://username:password@localhost:5672',
            method=connection.ConnectionType.URL,
            config=self.basic_config
        )
        self.basic_connection_host = connection.BasicConnection(
            connstr='localhost',
            method=connection.ConnectionType.HOST,
            username='username',
            password='password',
            config=self.basic_config
        )

    def test_base_name(self):
        self.assertEqual(connection.BasicConnection.__name__, 'BasicConnection')

    def test_base_initialization_url(self):
        self.assertEqual(self.basic_connection._connection_string, 'amqp://username:password@localhost:5672')
        self.assertEqual(self.basic_connection._method, connection.ConnectionType.URL)
        self.assertIsInstance(self.basic_connection._config, configs.BasicConfig)

    def test_base_initialization_host(self):
        self.basic_config = configs.BasicConfig()
        self.assertEqual(self.basic_connection_host._connection_string, 'localhost')
        self.assertEqual(self.basic_connection_host._method, connection.ConnectionType.HOST)
        self.assertEqual(self.basic_connection_host._username, 'username')
        self.assertEqual(self.basic_connection_host._password, 'password')
        self.assertIsInstance(self.basic_connection_host._config, configs.BasicConfig)

    def test__append_url_parameters(self):
        self.assertEqual(
            self.basic_connection._append_url_parameters(uri=self.basic_connection._connection_string),
            'amqp://username:password@localhost:5672?connection_attempts=3&heartbeat=3600&retry_delay=1'
        )

    def test_resolve_uri_parameters(self):
        self.assertEqual(
            self.basic_connection.resolve_uri_parameters(uri=self.basic_connection._connection_string),
            'amqp://username:password@localhost:5672?connection_attempts=3&heartbeat=3600&retry_delay=1'
        )

    def test_connect_parameters_and_nested_credentials_success(self):
        def mock_parameters(parameters):
            return parameters
        for connection_method in [
            {
                self.basic_connection: pika.URLParameters
            },
            {
                self.basic_connection_host: pika.ConnectionParameters
            }
        ]:
            with TA.patch(pika, 'BlockingConnection', mock_parameters):
                connection_parameters = [k for k, v in connection_method.items()][0].connect(poker_id='abc')
                self.assertIsInstance(connection_parameters, [v for k, v in connection_method.items()][0])
                self.assertEqual(connection_parameters._host, 'localhost')
                self.assertEqual(connection_parameters._credentials.username, 'username')
                self.assertEqual(connection_parameters._credentials.password, 'password')
                self.assertEqual(connection_parameters._connection_attempts, 3)
                self.assertEqual(connection_parameters._heartbeat, 3600)
                self.assertEqual(connection_parameters._retry_delay, 1)

    def test_connect_parameters_and_nested_credentials_failure(self):
        def mock_connection_error(parameters):
            raise Exception('Pokeman connecting to AMQP broker FAILED!')

        with TA.patch(pika, 'BlockingConnection', mock_connection_error):
            with self.assertRaisesRegex(Exception, 'Pokeman connecting to AMQP broker FAILED!'):
                self.basic_connection.connect(poker_id='abc')

    def test_disconnect_success(self):
        self.assertEqual(self.basic_connection.disconnect(connection=self.ConnectionMock(
            status='success'
        )), None)

    def test_disconnect_failure_connection_could_not_be_closed(self):
        def mock_connection_close_exception():
            raise Exception('Closing Pokeman connection FAILED!')

        with self.assertRaisesRegex(Exception, 'FAILED'):
            self.basic_connection.disconnect(connection=self.ConnectionMock(
                status='failure'
            )
            )

    def test_disconnect_failure_connection_not_is_closed(self):
        with self.assertRaisesRegex(ConnectionError, 'FAILED'):
            self.basic_connection.disconnect(connection=self.ConnectionMock(
                status='failure'
            )
            )


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod


class AbstractConfig(ABC):
    """
    Abstract base class for Enterprise Service Bus connection configuration.
    """
    @abstractmethod
    def __init__(
            self,
            connection_attempts,
            heartbeat,
            retry_delay
    ):
        self.CONNECTION_ATTEMPTS = connection_attempts
        self.HEARTBEAT = heartbeat
        self.RETRY_DELAY = retry_delay


class BasicConfig(AbstractConfig):
    """
    Basic configuration for Enterprise Service Bus connection.
    """
    def __init__(
            self,
            connection_attempts=3,
            heartbeat=3600,
            retry_delay=1
    ):
        super().__init__(
            connection_attempts=connection_attempts,
            heartbeat=heartbeat,
            retry_delay=retry_delay
        )



from .polling_consumer import PollingConsumer
from .basic_message_consumer import BasicMessageConsumer
from.selective_consumer import SelectiveConsumer

__all__ = ['PollingConsumer', 'BasicMessageConsumer', 'SelectiveConsumer']